/*
 *  Some utilities for writing and reading DVAVI files.
 *  These are not intended to serve for a full blown
 *  DVAVI handling software (this would be much too complex)
 *  The only intention is to write out MJPEG encoded
 *  DVAVIs with sound and to be able to read them back again.
 *  These utilities should work with other types of codecs too, however.
 *
 *  Copyright (C) 1999 Rainer Johanni <Rainer@Johanni.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "dvavilib.h"
#include "dvavi_struct.h"

/* The following variable indicates the kind of error */

long DVAVI_errno = 0;

/*******************************************************************
 *                                                                 *
 *    Utilities for writing an DVAVI File                            *
 *                                                                 *
 *******************************************************************/

/* DVAVI_MAX_LEN: The maximum length of an DVAVI file, we stay a bit below
    the 2GB limit (Remember: 2*10^9 is smaller than 2 GB) */

#define DVAVI_MAX_LEN 2000000000

/* HEADERBYTES: The number of bytes to reserve for the header */

#define HEADERBYTES 2048

#define PAD_EVEN(x) ( ((x)+1) & ~1 )


/* Copy n into dst as a 4 byte, little endian number.
   Should also work on big endian machines */

static void long2str(unsigned char *dst, int n)
{
   dst[0] = (n    )&0xff;
   dst[1] = (n>> 8)&0xff;
   dst[2] = (n>>16)&0xff;
   dst[3] = (n>>24)&0xff;
}

/* Convert a string of 4 or 2 bytes to a number,
   also working on big endian machines */

static unsigned long str2ulong(unsigned char *str)
{
   return ( str[0] | (str[1]<<8) | (str[2]<<16) | (str[3]<<24) );
}
static unsigned long str2ushort(unsigned char *str)
{
   return ( str[0] | (str[1]<<8) );
}

/* Calculate audio sample size from number of bits and number of channels.
   This may have to be adjusted for eg. 12 bits and stereo */

static int avi_sampsize(dvavi_t *DVAVI)
{
   int s;
   s = ((DVAVI->a_bits+7)/8)*DVAVI->a_chans;
   if(s==0) s=1; /* avoid possible zero divisions */
   return s;
}

/* Add a chunk (=tag and data) to the DVAVI file,
   returns -1 on write error, 0 on success */

static int avi_add_chunk(dvavi_t *DVAVI, unsigned char *tag, unsigned char *data, int length)
{
   unsigned char c[8];

   /* Copy tag and length int c, so that we need only 1 write system call
      for these two values */

   memcpy(c,tag,4);
   long2str(c+4,length);

   /* Output tag, length and data, restore previous position
      if the write fails */

   length = PAD_EVEN(length);

   if( write(DVAVI->fdes,c,8) != 8 ||
       write(DVAVI->fdes,data,length) != length )
   {
      lseek(DVAVI->fdes,DVAVI->pos,SEEK_SET);
      DVAVI_errno = DVAVI_ERR_WRITE;
      return -1;
   }

   /* Update file position */

   DVAVI->pos += 8 + length;

   return 0;
}

static int avi_add_index_entry(dvavi_t *DVAVI, unsigned char *tag, long flags, long pos, long len)
{
   void *ptr;

   if(DVAVI->n_idx>=DVAVI->max_idx)
   {
      ptr = realloc((void *)DVAVI->idx,(DVAVI->max_idx+4096)*16);
      if(ptr == 0)
      {
         DVAVI_errno = DVAVI_ERR_NO_MEM;
         return -1;
      }
      DVAVI->max_idx += 4096;
      DVAVI->idx = (unsigned char((*)[16]) ) ptr;
   }

   /* Add index entry */

   memcpy(DVAVI->idx[DVAVI->n_idx],tag,4);
   long2str(DVAVI->idx[DVAVI->n_idx]+ 4,flags);
   long2str(DVAVI->idx[DVAVI->n_idx]+ 8,pos);
   long2str(DVAVI->idx[DVAVI->n_idx]+12,len);

   /* Update counter */

   DVAVI->n_idx++;

   return 0;
}

/*
   DVAVI_open_output_file: Open an DVAVI File and write a bunch
                         of zero bytes as space for the header.

   returns a pointer to dvavi_t on success, a zero pointer on error
*/

dvavi_t* DVAVI_open_output_file(char * filename)
{
   dvavi_t *DVAVI;
   int i;
   unsigned char DVAVI_header[HEADERBYTES];

   /* Allocate the dvavi_t struct and zero it */

   DVAVI = (dvavi_t *) malloc(sizeof(dvavi_t));
   if(DVAVI==0)
   {
      DVAVI_errno = DVAVI_ERR_NO_MEM;
      return 0;
   }
   memset((void *)DVAVI,0,sizeof(dvavi_t));

   /* Since Linux needs a long time when deleting big files,
      we do not truncate the file when we open it.
      Instead it is truncated when the DVAVI file is closed */

   DVAVI->fdes = open(filename,O_RDWR|O_CREAT,0600);
   if (DVAVI->fdes < 0)
   {
      DVAVI_errno = DVAVI_ERR_OPEN;
      free(DVAVI);
      return 0;
   }

   /* Write out HEADERBYTES bytes, the header will go here
      when we are finished with writing */

   for (i=0;i<HEADERBYTES;i++) DVAVI_header[i] = 0;
   i = write(DVAVI->fdes,DVAVI_header,HEADERBYTES);
   if (i != HEADERBYTES)
   {
      close(DVAVI->fdes);
      DVAVI_errno = DVAVI_ERR_WRITE;
      free(DVAVI);
      return 0;
   }

   DVAVI->pos  = HEADERBYTES;
   DVAVI->mode = DVAVI_MODE_WRITE; /* open for writing */

   return DVAVI;
}

void DVAVI_set_video(dvavi_t *DVAVI, int width, int height, double fps, char *compressor)
{
   /* may only be called if file is open for writing */

   if(DVAVI->mode==DVAVI_MODE_READ) return;

   DVAVI->width  = width;
   DVAVI->height = height;
   DVAVI->fps    = fps;
   memcpy(DVAVI->compressor,compressor,4);
   DVAVI->compressor[4] = 0;

}

void DVAVI_set_audio(dvavi_t *DVAVI, int channels, long rate, int bits, int format)
{
   /* may only be called if file is open for writing */

   if(DVAVI->mode==DVAVI_MODE_READ) return;

   DVAVI->a_chans = channels;
   DVAVI->a_rate  = rate;
   DVAVI->a_bits  = bits;
   DVAVI->a_fmt   = format;

}

#define OUT4CC(s) \
   if(nhb<=HEADERBYTES-4) memcpy(DVAVI_header+nhb,s,4); nhb += 4

#define OUTLONG(n) \
   if(nhb<=HEADERBYTES-4) long2str(DVAVI_header+nhb,n); nhb += 4

#define OUTSHRT(n) \
   if(nhb<=HEADERBYTES-2) { \
      DVAVI_header[nhb  ] = (n   )&0xff; \
      DVAVI_header[nhb+1] = (n>>8)&0xff; \
   } \
   nhb += 2

/*
  Write the header of an DVAVI file and close it.
  returns 0 on success, -1 on write error.
*/

static int avi_close_output_file(dvavi_t *DVAVI)
{

   int ret, njunk, sampsize, hasIndex, ms_per_frame, idxerror, flag;
   int movi_len, hdrl_start, strl_start;
   unsigned char DVAVI_header[HEADERBYTES];
   long nhb;

   /* Calculate length of movi list */

   movi_len = DVAVI->pos - HEADERBYTES + 4;

   /* Try to ouput the index entries. This may fail e.g. if no space
      is left on device. We will report this as an error, but we still
      try to write the header correctly (so that the file still may be
      readable in the most cases */

   idxerror = 0;
   ret = avi_add_chunk(DVAVI,"idx1",(void*)DVAVI->idx,DVAVI->n_idx*16);
   hasIndex = (ret==0);
   if(ret)
   {
      idxerror = 1;
      DVAVI_errno = DVAVI_ERR_WRITE_INDEX;
   }

   /* Calculate Microseconds per frame */

   if(DVAVI->fps < 0.001)
      ms_per_frame = 0;
   else
      ms_per_frame = 1000000./DVAVI->fps + 0.5;

   /* Prepare the file header */

   nhb = 0;

   /* The RIFF header */

   OUT4CC ("RIFF");
   OUTLONG(DVAVI->pos - 8);    /* # of bytes to follow */
   OUT4CC ("DVAVI ");

   /* Start the header list */

   OUT4CC ("LIST");
   OUTLONG(0);        /* Length of list in bytes, don't know yet */
   hdrl_start = nhb;  /* Store start position */
   OUT4CC ("hdrl");

   /* The main DVAVI header */

   /* The Flags in DVAVI File header */

#define DVAVIF_HASINDEX           0x00000010      /* Index at end of file */
#define DVAVIF_MUSTUSEINDEX       0x00000020
#define DVAVIF_ISINTERLEAVED      0x00000100
#define DVAVIF_TRUSTCKTYPE        0x00000800      /* Use CKType to find key frames */
#define DVAVIF_WASCAPTUREFILE     0x00010000
#define DVAVIF_COPYRIGHTED        0x00020000

   OUT4CC ("avih");
   OUTLONG(56);                 /* # of bytes to follow */
   OUTLONG(ms_per_frame);       /* Microseconds per frame */
   OUTLONG(10000000);           /* MaxBytesPerSec, I hope this will never be used */
   OUTLONG(0);                  /* PaddingGranularity (whatever that might be) */
                                /* Other sources call it 'reserved' */
   flag = DVAVIF_WASCAPTUREFILE;
   if(hasIndex) flag |= DVAVIF_HASINDEX;
   if(hasIndex && DVAVI->must_use_index) flag |= DVAVIF_MUSTUSEINDEX;
   OUTLONG(flag);               /* Flags */
   OUTLONG(DVAVI->video_frames);  /* TotalFrames */
   OUTLONG(0);                  /* InitialFrames */
   if (DVAVI->audio_bytes)
      { OUTLONG(2); }           /* Streams */
   else
      { OUTLONG(1); }           /* Streams */
   OUTLONG(0);                  /* SuggestedBufferSize */
   OUTLONG(DVAVI->width);         /* Width */
   OUTLONG(DVAVI->height);        /* Height */
                                /* MS calls the following 'reserved': */
   OUTLONG(0);                  /* TimeScale:  Unit used to measure time */
   OUTLONG(0);                  /* DataRate:   Data rate of playback     */
   OUTLONG(0);                  /* StartTime:  Starting time of DVAVI data */
   OUTLONG(0);                  /* DataLength: Size of DVAVI data chunk    */


   /* Start the video stream list ---------------------------------- */

   OUT4CC ("LIST");
   OUTLONG(0);        /* Length of list in bytes, don't know yet */
   strl_start = nhb;  /* Store start position */
   OUT4CC ("strl");

   /* The video stream header */

   OUT4CC ("strh");
   OUTLONG(64);                 /* # of bytes to follow */
   OUT4CC ("vids");             /* Type */
   OUT4CC (DVAVI->compressor);    /* Handler */
   OUTLONG(0);                  /* Flags */
   OUTLONG(0);                  /* Reserved, MS says: wPriority, wLanguage */
   OUTLONG(0);                  /* InitialFrames */
   OUTLONG(ms_per_frame);       /* Scale */
   OUTLONG(1000000);            /* Rate: Rate/Scale == samples/second */
   OUTLONG(0);                  /* Start */
   OUTLONG(DVAVI->video_frames);  /* Length */
   OUTLONG(0);                  /* SuggestedBufferSize */
   OUTLONG(-1);                 /* Quality */
   OUTLONG(0);                  /* SampleSize */
   OUTLONG(0);                  /* Frame */
   OUTLONG(0);                  /* Frame */
   OUTLONG(0);                  /* Frame */
   OUTLONG(0);                  /* Frame */

   /* The video stream format */

   OUT4CC ("strf");
   OUTLONG(40);                 /* # of bytes to follow */
   OUTLONG(40);                 /* Size */
   OUTLONG(DVAVI->width);         /* Width */
   OUTLONG(DVAVI->height);        /* Height */
   OUTSHRT(1); OUTSHRT(24);     /* Planes, Count */
   OUT4CC (DVAVI->compressor);    /* Compression */
   OUTLONG(DVAVI->width*DVAVI->height);  /* SizeImage (in bytes?) */
   OUTLONG(0);                  /* XPelsPerMeter */
   OUTLONG(0);                  /* YPelsPerMeter */
   OUTLONG(0);                  /* ClrUsed: Number of colors used */
   OUTLONG(0);                  /* ClrImportant: Number of colors important */

   /* Finish stream list, i.e. put number of bytes in the list to proper pos */

   long2str(DVAVI_header+strl_start-4,nhb-strl_start);

   if (DVAVI->a_chans && DVAVI->audio_bytes)
   {

   sampsize = avi_sampsize(DVAVI);

   /* Start the audio stream list ---------------------------------- */

   OUT4CC ("LIST");
   OUTLONG(0);        /* Length of list in bytes, don't know yet */
   strl_start = nhb;  /* Store start position */
   OUT4CC ("strl");

   /* The audio stream header */

   OUT4CC ("strh");
   OUTLONG(64);            /* # of bytes to follow */
   OUT4CC ("auds");
   OUT4CC ("\0\0\0\0");
   OUTLONG(0);             /* Flags */
   OUTLONG(0);             /* Reserved, MS says: wPriority, wLanguage */
   OUTLONG(0);             /* InitialFrames */
   OUTLONG(sampsize);      /* Scale */
   OUTLONG(sampsize*DVAVI->a_rate); /* Rate: Rate/Scale == samples/second */
   OUTLONG(0);             /* Start */
   OUTLONG(DVAVI->audio_bytes/sampsize);   /* Length */
   OUTLONG(0);             /* SuggestedBufferSize */
   OUTLONG(-1);            /* Quality */
   OUTLONG(sampsize);      /* SampleSize */
   OUTLONG(0);             /* Frame */
   OUTLONG(0);             /* Frame */
   OUTLONG(0);             /* Frame */
   OUTLONG(0);             /* Frame */

   /* The audio stream format */

   OUT4CC ("strf");
   OUTLONG(16);                   /* # of bytes to follow */
   OUTSHRT(DVAVI->a_fmt);           /* Format */
   OUTSHRT(DVAVI->a_chans);         /* Number of channels */
   OUTLONG(DVAVI->a_rate);          /* SamplesPerSec */
   OUTLONG(sampsize*DVAVI->a_rate); /* AvgBytesPerSec */
   OUTSHRT(sampsize);             /* BlockAlign */
   OUTSHRT(DVAVI->a_bits);          /* BitsPerSample */

   /* Finish stream list, i.e. put number of bytes in the list to proper pos */

   long2str(DVAVI_header+strl_start-4,nhb-strl_start);

   }

   /* Finish header list */

   long2str(DVAVI_header+hdrl_start-4,nhb-hdrl_start);

   /* Calculate the needed amount of junk bytes, output junk */

   njunk = HEADERBYTES - nhb - 8 - 12;

   /* Safety first: if njunk <= 0, somebody has played with
      HEADERBYTES without knowing what (s)he did.
      This is a fatal error */

   if(njunk<=0)
   {
      fprintf(stderr,"DVAVI_close_output_file: # of header bytes too small\n");
      exit(1);
   }

   OUT4CC ("JUNK");
   OUTLONG(njunk);
   memset(DVAVI_header+nhb,0,njunk);
   nhb += njunk;

   /* Start the movi list */

   OUT4CC ("LIST");
   OUTLONG(movi_len); /* Length of list in bytes */
   OUT4CC ("movi");

   /* Output the header, truncate the file to the number of bytes
      actually written, report an error if someting goes wrong */

   if ( lseek(DVAVI->fdes,0,SEEK_SET)<0 ||
        write(DVAVI->fdes,DVAVI_header,HEADERBYTES)!=HEADERBYTES ||
        ftruncate(DVAVI->fdes,DVAVI->pos)<0 )
   {
      DVAVI_errno = DVAVI_ERR_CLOSE;
      return -1;
   }

   if(idxerror) return -1;

   return 0;
}
/*
   DVAVI_write_data:
   Add video or audio data to the file;

   Return values:
    0    No error;
   -1    Error, DVAVI_errno is set appropriatly;

*/

static int avi_write_data(dvavi_t *DVAVI, char *data, long length, int audio)
{
   int n;

   /* Check for maximum file length */

   if ( (DVAVI->pos + 8 + length + 8 + (DVAVI->n_idx+1)*16) > DVAVI_MAX_LEN )
   {
      DVAVI_errno = DVAVI_ERR_SIZELIM;
      return -1;
   }

   /* Add index entry */

   if(audio)
      n = avi_add_index_entry(DVAVI,"01wb",0x00,DVAVI->pos,length);
   else
      n = avi_add_index_entry(DVAVI,"00db",0x10,DVAVI->pos,length);

   if(n) return -1;

   /* Output tag and data */

   if(audio)
      n = avi_add_chunk(DVAVI,"01wb",data,length);
   else
      n = avi_add_chunk(DVAVI,"00db",data,length);

   if (n) return -1;

   return 0;
}

int DVAVI_write_frame(dvavi_t *DVAVI, char *data, long bytes)
{
   long pos;

   if(DVAVI->mode==DVAVI_MODE_READ) { DVAVI_errno = DVAVI_ERR_NOT_PERM; return -1; }

   pos = DVAVI->pos;
   if( avi_write_data(DVAVI,data,bytes,0) ) return -1;
   DVAVI->last_pos = pos;
   DVAVI->last_len = bytes;
   DVAVI->video_frames++;
   return 0;
}

int DVAVI_dup_frame(dvavi_t *DVAVI)
{
   if(DVAVI->mode==DVAVI_MODE_READ) { DVAVI_errno = DVAVI_ERR_NOT_PERM; return -1; }

   if(DVAVI->last_pos==0) return 0; /* No previous real frame */
   if(avi_add_index_entry(DVAVI,"00db",0x10,DVAVI->last_pos,DVAVI->last_len)) return -1;
   DVAVI->video_frames++;
   DVAVI->must_use_index = 1;
   return 0;
}

int DVAVI_write_audio(dvavi_t *DVAVI, char *data, long bytes)
{
   if(DVAVI->mode==DVAVI_MODE_READ) { DVAVI_errno = DVAVI_ERR_NOT_PERM; return -1; }

   if( avi_write_data(DVAVI,data,bytes,1) ) return -1;
   DVAVI->audio_bytes += bytes;
   return 0;
}

long DVAVI_bytes_remain(dvavi_t *DVAVI)
{
   if(DVAVI->mode==DVAVI_MODE_READ) return 0;

   return ( DVAVI_MAX_LEN - (DVAVI->pos + 8 + 16*DVAVI->n_idx));
}

/*******************************************************************
 *                                                                 *
 *    Utilities for reading video and audio from an DVAVI File       *
 *                                                                 *
 *******************************************************************/

int DVAVI_close(dvavi_t *DVAVI)
{
   int ret;

   /* If the file was open for writing, the header and index still have
      to be written */

   if(DVAVI->mode == DVAVI_MODE_WRITE)
      ret = avi_close_output_file(DVAVI);
   else
      ret = 0;

   /* Even if there happened a error, we first clean up */

   close(DVAVI->fdes);
   if(DVAVI->idx) free(DVAVI->idx);
   if(DVAVI->video_index) free(DVAVI->video_index);
   if(DVAVI->audio_index) free(DVAVI->audio_index);
   free(DVAVI);

   return ret;
}


#define ERR_EXIT(x) \
{ \
   DVAVI_close(DVAVI); \
   DVAVI_errno = x; \
   return 0; \
}

dvavi_t *DVAVI_open_input_file(char *filename, int getIndex)
{
   dvavi_t *DVAVI;
   long i, n, rate, scale, idx_type;
   unsigned char *hdrl_data;
   long hdrl_len;
   long nvi, nai, ioff;
   long tot;
   int lasttag = 0;
   int vids_strh_seen = 0;
   int vids_strf_seen = 0;
   int auds_strh_seen = 0;
   int auds_strf_seen = 0;
   int num_stream = 0;
   char data[256];

   /* Create dvavi_t structure */

   DVAVI = (dvavi_t *) malloc(sizeof(dvavi_t));
   if(DVAVI==NULL)
   {
      DVAVI_errno = DVAVI_ERR_NO_MEM;
      return 0;
   }
   memset((void *)DVAVI,0,sizeof(dvavi_t));

   DVAVI->mode = DVAVI_MODE_READ; /* open for reading */

   /* Open the file */

   DVAVI->fdes = open(filename,O_RDONLY);
   if(DVAVI->fdes < 0)
   {
      DVAVI_errno = DVAVI_ERR_OPEN;
      free(DVAVI);
      return 0;
   }

   /* Read first 12 bytes and check that this is an DVAVI file */

   if( read(DVAVI->fdes,data,12) != 12 ) ERR_EXIT(DVAVI_ERR_READ)

   if( strncasecmp(data  ,"RIFF",4) !=0 ||
       strncasecmp(data+8,"AVI ",4) !=0 ) ERR_EXIT(DVAVI_ERR_NO_DVAVI)

   /* Go through the DVAVI file and extract the header list,
      the start position of the 'movi' list and an optionally
      present idx1 tag */

   hdrl_data = 0;

   while(1)
   {
      if( read(DVAVI->fdes,data,8) != 8 ) break; /* We assume it's EOF */

      n = str2ulong(data+4);
      n = PAD_EVEN(n);

      if(strncasecmp(data,"LIST",4) == 0)
      {
         if( read(DVAVI->fdes,data,4) != 4 ) ERR_EXIT(DVAVI_ERR_READ)
         n -= 4;
         if(strncasecmp(data,"hdrl",4) == 0)
         {
            hdrl_len = n;
            hdrl_data = (unsigned char *) malloc(n);
            if(hdrl_data==0) ERR_EXIT(DVAVI_ERR_NO_MEM)
            if( read(DVAVI->fdes,hdrl_data,n) != n ) ERR_EXIT(DVAVI_ERR_READ)
         }
         else if(strncasecmp(data,"movi",4) == 0)
         {
            DVAVI->movi_start = lseek(DVAVI->fdes,0,SEEK_CUR);
            lseek(DVAVI->fdes,n,SEEK_CUR);
         }
         else
            lseek(DVAVI->fdes,n,SEEK_CUR);
      }
      else if(strncasecmp(data,"idx1",4) == 0)
      {
         /* n must be a multiple of 16, but the reading does not
            break if this is not the case */

         DVAVI->n_idx = DVAVI->max_idx = n/16;
         DVAVI->idx = (unsigned  char((*)[16]) ) malloc(n);
         if(DVAVI->idx==0) ERR_EXIT(DVAVI_ERR_NO_MEM)
         if( read(DVAVI->fdes,DVAVI->idx,n) != n ) ERR_EXIT(DVAVI_ERR_READ)
      }
      else
         lseek(DVAVI->fdes,n,SEEK_CUR);
   }

   if(!hdrl_data      ) ERR_EXIT(DVAVI_ERR_NO_HDRL)
   if(!DVAVI->movi_start) ERR_EXIT(DVAVI_ERR_NO_MOVI)

   /* Interpret the header list */
   memcpy(DVAVI->compressor,"WHAT",4);
   DVAVI->compressor[4] = 0;
   DVAVI->video_strn = 0;
   DVAVI->audio_strn = 1;     

   for(i=0;i<hdrl_len;)
   {
      /* List tags are completly ignored */

      if(strncasecmp(hdrl_data+i,"LIST",4)==0) { i+= 12; continue; }

      n = str2ulong(hdrl_data+i+4);
      n = PAD_EVEN(n);

      /* Interpret the tag and its args */

      if(strncasecmp(hdrl_data+i,"strh",4)==0)
      {
		AVIStreamHeader *a;
        i += 8;
        a=(AVIStreamHeader *)(hdrl_data+i);
		 if(strncasecmp(hdrl_data+i,"vids",4)==0) {
         	scale = a->dwScale;
         	rate  = a->dwRate;
         	if(scale!=0) DVAVI->fps = (double)rate/(double)scale;
		} else if(strncasecmp(hdrl_data+i,"auds",4)==0) {
            DVAVI->audio_bytes=a->dwLength;
		}
      }
      else if(strncasecmp(hdrl_data+i,"strf",4)==0)
      {
			WAVEFORMATEX *w;
			i+=8;
			w=(WAVEFORMATEX *)(hdrl_data+i);
         	DVAVI->a_chans = w->nChannels;
         	DVAVI->a_bits  = w->wBitsPerSample;
/*            DVAVI->audio_bytes = str2ulong(hdrl_data+i+32)*avi_sampsize(DVAVI);*/
            DVAVI->a_fmt   = w->wFormatTag;
            DVAVI->a_rate  = w->nSamplesPerSec;
      }
      else if(strncasecmp(hdrl_data+i,"avih",4)==0) {
		MainAVIHeader *a;
        i += 8;
		a=(MainAVIHeader *)(hdrl_data+i);
        DVAVI->video_frames = a->dwTotalFrames;
        DVAVI->width  = a->dwWidth;
        DVAVI->height = a->dwHeight;
	  }
      else if(strncasecmp(hdrl_data+i,"dmlh",4)==0) {
         i += 8;
		/* dmlh can be ignored */
	  }
      else if(strncasecmp(hdrl_data+i,"indx",4)==0) {
         i += 8;
		printf("handling indx!\n");
	  }
	  else 
	  {
			printf("unknown chunk!  ");fflush(stdout);
			write(1,hdrl_data+i,4);
			printf("\n");
			
         	i += 8;
      }

      i += n;
   }
   free(hdrl_data);

	/* we find the true audio bytes here */
    /* wait! seems like it's already okay... */
#if 0
   printf("audio length: %d\n",DVAVI->audio_bytes); 
   DVAVI->audio_bytes*=DVAVI->a_rate;
#endif

   printf("Compressor: %s\n",DVAVI->compressor);
   printf("fps: %f\n",DVAVI->fps);
   printf("video_frames: %d\n",DVAVI->video_frames);
   printf("video_strn: %d\n",DVAVI->video_strn);
   printf("audio_bytes: %d\n",DVAVI->audio_bytes);
   printf("audio_strn: %d\n",DVAVI->audio_strn);
   printf("width: %d\n",DVAVI->width);
   printf("height: %d\n",DVAVI->height);
   printf("a_fmt: %d\n",DVAVI->a_fmt);
   printf("a_chans: %d\n",DVAVI->a_chans);
   printf("a_rate: %d\n",DVAVI->a_rate);
   printf("a_bits: %d\n",DVAVI->a_bits);


   if(DVAVI->video_frames==0) ERR_EXIT(DVAVI_ERR_NO_VIDS) 

   DVAVI->video_tag[0] = DVAVI->video_strn/10 + '0';
   DVAVI->video_tag[1] = DVAVI->video_strn%10 + '0';
   DVAVI->video_tag[2] = 'd';
   DVAVI->video_tag[3] = 'c';

   /* Audio tag is set to "99wb" if no audio present */
   if(!DVAVI->a_chans) DVAVI->audio_strn = 99;

   DVAVI->audio_tag[0] = DVAVI->audio_strn/10 + '0';
   DVAVI->audio_tag[1] = DVAVI->audio_strn%10 + '0';
   DVAVI->audio_tag[2] = 'w';
   DVAVI->audio_tag[3] = 'b';

   lseek(DVAVI->fdes,DVAVI->movi_start,SEEK_SET);

   /* get index if wanted */

   if(!getIndex) return DVAVI;

   /* if the file has an idx1, check if this is relative
      to the start of the file or to the start of the movi list */

   idx_type = 0;

   if(DVAVI->idx)
   {
      long pos, len;

      /* Search the first videoframe in the idx1 and look where
         it is in the file */

      for(i=0;i<DVAVI->n_idx;i++)
         if( strncasecmp(DVAVI->idx[i],DVAVI->video_tag,3)==0 ) break;
      if(i>=DVAVI->n_idx) ERR_EXIT(DVAVI_ERR_NO_VIDS)

      pos = str2ulong(DVAVI->idx[i]+ 8);
      len = str2ulong(DVAVI->idx[i]+12);

      lseek(DVAVI->fdes,pos,SEEK_SET);
      if(read(DVAVI->fdes,data,8)!=8) ERR_EXIT(DVAVI_ERR_READ)
      if( strncasecmp(data,DVAVI->idx[i],4)==0 && str2ulong(data+4)==len )
      {
         idx_type = 1; /* Index from start of file */
      }
      else
      {
         lseek(DVAVI->fdes,pos+DVAVI->movi_start-4,SEEK_SET);
         if(read(DVAVI->fdes,data,8)!=8) ERR_EXIT(DVAVI_ERR_READ)
         if( strncasecmp(data,DVAVI->idx[i],4)==0 && str2ulong(data+4)==len )
         {
            idx_type = 2; /* Index from start of movi list */
         }
      }
      /* idx_type remains 0 if neither of the two tests above succeeds */
   }

   if(idx_type == 0)
   {
      /* we must search through the file to get the index */

      lseek(DVAVI->fdes, DVAVI->movi_start, SEEK_SET);

      DVAVI->n_idx = 0;

      while(1)
      {
         if( read(DVAVI->fdes,data,8) != 8 ) break;
         n = str2ulong(data+4);

         /* The movi list may contain sub-lists, ignore them */

         if(strncasecmp(data,"LIST",4)==0)
         {
            lseek(DVAVI->fdes,4,SEEK_CUR);
            continue;
         }

         /* Check if we got a tag ##db, ##dc or ##wb */

         if( ( (data[2]=='d' || data[2]=='D') &&
               (data[3]=='b' || data[3]=='B' || data[3]=='c' || data[3]=='C') )
          || ( (data[2]=='w' || data[2]=='W') &&
               (data[3]=='b' || data[3]=='B') ) )
         {
            avi_add_index_entry(DVAVI,data,0,lseek(DVAVI->fdes,0,SEEK_CUR)-8,n);
         }

         lseek(DVAVI->fdes,PAD_EVEN(n),SEEK_CUR);
      }
      idx_type = 1;
   }

   /* Now generate the video index and audio index arrays */

   nvi = 0;
   nai = 0;

   for(i=0;i<DVAVI->n_idx;i++)
   {
      if(strncasecmp(DVAVI->idx[i],DVAVI->video_tag,3) == 0) nvi++;
      if(strncasecmp(DVAVI->idx[i],DVAVI->audio_tag,4) == 0) nai++;
   }

   DVAVI->video_frames = nvi;
   DVAVI->audio_chunks = nai;

   if(DVAVI->video_frames==0) ERR_EXIT(DVAVI_ERR_NO_VIDS)
   DVAVI->video_index = (dvvideo_index_entry *) malloc(nvi*sizeof(dvvideo_index_entry));
   if(DVAVI->video_index==0) ERR_EXIT(DVAVI_ERR_NO_MEM)
   if(DVAVI->audio_chunks)
   {
      DVAVI->audio_index = (dvaudio_index_entry *) malloc(nai*sizeof(dvaudio_index_entry));
      if(DVAVI->audio_index==0) ERR_EXIT(DVAVI_ERR_NO_MEM)
   }

   nvi = 0;
   nai = 0;
   tot = 0;
   ioff = idx_type == 1 ? 8 : DVAVI->movi_start+4;

   for(i=0;i<DVAVI->n_idx;i++)
   {
      if(strncasecmp(DVAVI->idx[i],DVAVI->video_tag,3) == 0)
      {
         DVAVI->video_index[nvi].pos = str2ulong(DVAVI->idx[i]+ 8)+ioff;
         DVAVI->video_index[nvi].len = str2ulong(DVAVI->idx[i]+12);
         nvi++;
      }
      if(strncasecmp(DVAVI->idx[i],DVAVI->audio_tag,4) == 0)
      {
         DVAVI->audio_index[nai].pos = str2ulong(DVAVI->idx[i]+ 8)+ioff;
         DVAVI->audio_index[nai].len = str2ulong(DVAVI->idx[i]+12);
         DVAVI->audio_index[nai].tot = tot;
         tot += DVAVI->audio_index[nai].len;
         nai++;
      }
   }

   DVAVI->audio_bytes = tot;

   /* Reposition the file */

   lseek(DVAVI->fdes,DVAVI->movi_start,SEEK_SET);
   DVAVI->video_pos = 0;

   return DVAVI;
}

long DVAVI_video_frames(dvavi_t *DVAVI)
{
   return DVAVI->video_frames;
}
int  DVAVI_video_width(dvavi_t *DVAVI)
{
   return DVAVI->width;
}
int  DVAVI_video_height(dvavi_t *DVAVI)
{
   return DVAVI->height;
}
double DVAVI_frame_rate(dvavi_t *DVAVI)
{
   return DVAVI->fps;
}
char* DVAVI_video_compressor(dvavi_t *DVAVI)
{
   return DVAVI->compressor;
}

int DVAVI_audio_channels(dvavi_t *DVAVI)
{
   return DVAVI->a_chans;
}
int DVAVI_audio_bits(dvavi_t *DVAVI)
{
   return DVAVI->a_bits;
}
int DVAVI_audio_format(dvavi_t *DVAVI)
{
   return DVAVI->a_fmt;
}
long DVAVI_audio_rate(dvavi_t *DVAVI)
{
   return DVAVI->a_rate;
}
long DVAVI_audio_bytes(dvavi_t *DVAVI)
{
   return DVAVI->audio_bytes;
}

long DVAVI_frame_size(dvavi_t *DVAVI, long frame)
{
   if(DVAVI->mode==DVAVI_MODE_WRITE) { DVAVI_errno = DVAVI_ERR_NOT_PERM; return -1; }
   if(!DVAVI->video_index)         { DVAVI_errno = DVAVI_ERR_NO_IDX;   return -1; }

   if(frame < 0 || frame >= DVAVI->video_frames) return 0;
   return(DVAVI->video_index[frame].len);
}

int DVAVI_seek_start(dvavi_t *DVAVI)
{
   if(DVAVI->mode==DVAVI_MODE_WRITE) { DVAVI_errno = DVAVI_ERR_NOT_PERM; return -1; }

   lseek(DVAVI->fdes,DVAVI->movi_start,SEEK_SET);
   DVAVI->video_pos = 0;
   return 0;
}

int DVAVI_set_video_position(dvavi_t *DVAVI, long frame)
{
   if(DVAVI->mode==DVAVI_MODE_WRITE) { DVAVI_errno = DVAVI_ERR_NOT_PERM; return -1; }
   if(!DVAVI->video_index)         { DVAVI_errno = DVAVI_ERR_NO_IDX;   return -1; }

   if (frame < 0 ) frame = 0;
   DVAVI->video_pos = frame;
   return 0;
}
      

long DVAVI_read_frame(dvavi_t *DVAVI, char *vidbuf)
{
   long n;

   if(DVAVI->mode==DVAVI_MODE_WRITE) { DVAVI_errno = DVAVI_ERR_NOT_PERM; return -1; }
   if(!DVAVI->video_index)         { DVAVI_errno = DVAVI_ERR_NO_IDX;   return -1; }

   if(DVAVI->video_pos < 0 || DVAVI->video_pos >= DVAVI->video_frames) return 0;
   n = DVAVI->video_index[DVAVI->video_pos].len;

   lseek(DVAVI->fdes, DVAVI->video_index[DVAVI->video_pos].pos, SEEK_SET);
   if (read(DVAVI->fdes,vidbuf,n) != n)
   {
      DVAVI_errno = DVAVI_ERR_READ;
      return -1;
   }

   DVAVI->video_pos++;

   return n;
}

int DVAVI_set_audio_position(dvavi_t *DVAVI, long byte)
{
   long n0, n1, n;

   if(DVAVI->mode==DVAVI_MODE_WRITE) { DVAVI_errno = DVAVI_ERR_NOT_PERM; return -1; }
   if(!DVAVI->audio_index)         { DVAVI_errno = DVAVI_ERR_NO_IDX;   return -1; }

   if(byte < 0) byte = 0;

   /* Binary search in the audio chunks */

   n0 = 0;
   n1 = DVAVI->audio_chunks;

   while(n0<n1-1)
   {
      n = (n0+n1)/2;
      if(DVAVI->audio_index[n].tot>byte)
         n1 = n;
      else
         n0 = n;
   }

   DVAVI->audio_posc = n0;
   DVAVI->audio_posb = byte - DVAVI->audio_index[n0].tot;

   return 0;
}

long DVAVI_read_audio(dvavi_t *DVAVI, char *audbuf, long bytes)
{
   long nr, pos, left, todo;

   if(DVAVI->mode==DVAVI_MODE_WRITE) { DVAVI_errno = DVAVI_ERR_NOT_PERM; return -1; }
   if(!DVAVI->audio_index)         { DVAVI_errno = DVAVI_ERR_NO_IDX;   return -1; }

   nr = 0; /* total number of bytes read */

   while(bytes>0)
   {
      left = DVAVI->audio_index[DVAVI->audio_posc].len - DVAVI->audio_posb;
      if(left==0)
      {
         if(DVAVI->audio_posc>=DVAVI->audio_chunks-1) return nr;
         DVAVI->audio_posc++;
         DVAVI->audio_posb = 0;
         continue;
      }
      if(bytes<left)
         todo = bytes;
      else
         todo = left;
      pos = DVAVI->audio_index[DVAVI->audio_posc].pos + DVAVI->audio_posb;
      lseek(DVAVI->fdes, pos, SEEK_SET);
      if (read(DVAVI->fdes,audbuf+nr,todo) != todo)
      {
         DVAVI_errno = DVAVI_ERR_READ;
         return -1;
      }
      bytes -= todo;
      nr    += todo;
      DVAVI->audio_posb += todo;
   }

   return nr;
}

/* DVAVI_read_data: Special routine for reading the next audio or video chunk
                  without having an index of the file. */

int DVAVI_read_data(dvavi_t *DVAVI, char *vidbuf, long max_vidbuf,
                              char *audbuf, long max_audbuf,
                              long *len)
{

/*
 * Return codes:
 *
 *    1 = video data read
 *    2 = audio data read
 *    0 = reached EOF
 *   -1 = video buffer too small
 *   -2 = audio buffer too small
 */

   int n;
   char data[8];

   if(DVAVI->mode==DVAVI_MODE_WRITE) return 0;

   while(1)
   {
      /* Read tag and length */

      if( read(DVAVI->fdes,data,8) != 8 ) return 0;

      /* if we got a list tag, ignore it */

      if(strncasecmp(data,"LIST",4) == 0)
      {
         lseek(DVAVI->fdes,4,SEEK_CUR);
         continue;
      }

      n = PAD_EVEN(str2ulong(data+4));

      if(strncasecmp(data,DVAVI->video_tag,3) == 0)
      {
         *len = n;
         DVAVI->video_pos++;
         if(n>max_vidbuf)
         {
            lseek(DVAVI->fdes,n,SEEK_CUR);
            return -1;
         }
         if(read(DVAVI->fdes,vidbuf,n) != n ) return 0;
         return 1;
      }
      else if(strncasecmp(data,DVAVI->audio_tag,4) == 0)
      {
         *len = n;
         if(n>max_audbuf)
         {
            lseek(DVAVI->fdes,n,SEEK_CUR);
            return -2;
         }
         if(read(DVAVI->fdes,audbuf,n) != n ) return 0;
         return 2;
         break;
      }
      else
         if(lseek(DVAVI->fdes,n,SEEK_CUR)<0)  return 0;
   }
}

/* DVAVI_print_error: Print most recent error (similar to perror) */

char *(dvavi_errors[]) =
{
  /*  0 */ "avilib - No Error",
  /*  1 */ "avilib - DVAVI file size limit reached",
  /*  2 */ "avilib - Error opening DVAVI file",
  /*  3 */ "avilib - Error reading from DVAVI file",
  /*  4 */ "avilib - Error writing to DVAVI file",
  /*  5 */ "avilib - Error writing index (file may still be useable)",
  /*  6 */ "avilib - Error closing DVAVI file",
  /*  7 */ "avilib - Operation (read/write) not permitted",
  /*  8 */ "avilib - Out of memory (malloc failed)",
  /*  9 */ "avilib - Not an DVAVI file",
  /* 10 */ "avilib - DVAVI file has no header list (corrupted?)",
  /* 11 */ "avilib - DVAVI file has no MOVI list (corrupted?)",
  /* 12 */ "avilib - DVAVI file has no video data",
  /* 13 */ "avilib - operation needs an index",
  /* 14 */ "avilib - Unkown Error"
};
static int num_dvavi_errors = sizeof(dvavi_errors)/sizeof(char*);

static char error_string[4096];

void DVAVI_print_error(char *str)
{
   int aerrno;

   aerrno = (DVAVI_errno>=0 && DVAVI_errno<num_dvavi_errors) ? DVAVI_errno : num_dvavi_errors-1;

   fprintf(stderr,"%s: %s\n",str,dvavi_errors[aerrno]);

   /* for the following errors, perror should report a more detailed reason: */

   if(DVAVI_errno == DVAVI_ERR_OPEN ||
      DVAVI_errno == DVAVI_ERR_READ ||
      DVAVI_errno == DVAVI_ERR_WRITE ||
      DVAVI_errno == DVAVI_ERR_WRITE_INDEX ||
      DVAVI_errno == DVAVI_ERR_CLOSE )
   {
      perror("REASON");
   }
}

char *DVAVI_strerror()
{
   int aerrno;

   aerrno = (DVAVI_errno>=0 && DVAVI_errno<num_dvavi_errors) ? DVAVI_errno : num_dvavi_errors-1;

   if(DVAVI_errno == DVAVI_ERR_OPEN ||
      DVAVI_errno == DVAVI_ERR_READ ||
      DVAVI_errno == DVAVI_ERR_WRITE ||
      DVAVI_errno == DVAVI_ERR_WRITE_INDEX ||
      DVAVI_errno == DVAVI_ERR_CLOSE )
   {
      sprintf(error_string,"%s - %s",dvavi_errors[aerrno],strerror(errno));
      return error_string;
   }
   else
   {
      return dvavi_errors[aerrno];
   }
}
